use std::f64::consts::PI;

use crate::language::Rational;

pub fn approximate_rational_cubic(b: &Rational, c: &Rational, d: &Rational, limit: u64) -> Vec<Rational> {
	let numerical_sols = solve_cubic_numerically(
		1.0,
		(b.num as f64) / (b.denom as f64),
		(c.num as f64) / (c.denom as f64),
		(d.num as f64) / (d.denom as f64));

	numerical_sols.into_iter().map(|x|rational_approx(x, limit)).collect()
}

// assuming leading coefficient is not 0, and also that we don't have a triple root
pub fn solve_cubic_numerically(a: f64, b: f64, c: f64, d: f64) -> Vec<f64> {
	assert_ne!(a, 0.0);

	let b = b/a;
	let c = c/a;
	let d = d/a;

	let b2 = b * b;
	let b3 = b2 * b;

	solve_depressed_cubic_numerically(
		c - b2 / 3.0,
		2.0 * b3 / 27.0 - b * c / 3.0 + d
	).into_iter().map(|u|u - b / 3.0 / a).collect()
}

fn solve_depressed_cubic_numerically(p: f64, q: f64) -> Vec<f64> {
	let disc = p * p * p / 27.0 + q * q / 4.0;

	if disc < 0.0 {
		let r = 2.0 * (-p / 3.0).sqrt();
		let t = 3.0 * q / 2.0 / p * (- 3.0 / p).sqrt();
		let phi = t.acos() / 3.0;

		vec![r * (phi).cos(),
			 r * (phi + 2.0 * PI / 3.0).cos(),
			 r * (phi + 4.0 * PI / 3.0).cos()]
	} else {
		// not implemented at the moment
		let u = - q / 2.0 + disc.sqrt();
		let v = - q / 2.0 - disc.sqrt();
		vec![u.powf(1.0/3.0) + v.powf(1.0/3.0)]
	}
}

// pretty inefficient implementation of continued fraction approximation
// but doesn't matter here
pub fn rational_approx(x: f64, limit: u64) -> Rational {
	if x < 0.0 {
		let Rational { num, denom } = rational_approx(-x, limit);
		return Rational { num: -num, denom }
	}

	let mut num = 0;
	let mut denom = 0;
	for l in 0 .. 10 {
		let Some((p,q)) = rational_approx_level(x, l) else { break; };

		if q > limit {
			break;
		}

		(num, denom) = (p, q);
	}

	Rational{
		num: num as i64,
		denom
	}
}

fn rational_approx_level(x: f64, level: usize) -> Option<(u64, u64)> {
	// we expect very big numbers or infinity if the previous iteration was exact
	if x > 1e9 {
		return None;
	}

	if level == 0 {
		Some((x as u64, 1))
	} else {
		let (p,q) = rational_approx_level(1.0 / (x - x.floor()), level-1)?;
		let floorx = x as u64;

		Some((q + floorx * p, p))
	}
}
