use egg::{AstSize, Extractor, Id, RecExpr, Runner};
use solveq::cubic::approximate_rational_cubic;
use solveq::language::{EquationLanguage, Rational, RULES, EGraph};
use solveq::normal_form::extract_normal_form;
use solveq::parse::parse_equation;
use solveq::output::print_term;

static TEST_EQUATIONS: &[&str] = &[
	"(x + 50) * 10 - 150 = 100",
	"(x - 2) * (x + 2) = 0",
	"x ^ 2 = 4",
	"x ^ 2 - 2 = 0",
	"x ^ 2 = 2 * x + 15",
	"(x ^ 2 - 2 * x - 15) * (x + 5) = 0",
	"x ^ 3 + 3 * x ^ 2 - 25 * x - 75 = 0",
	"x ^ 3 - 91 * x  - 90 = 0",
];

fn main() {
	for eq in TEST_EQUATIONS {
		println!("Equation: {}", *eq);

		let solutions = solve(*eq, true);

		let solutions_str: Vec<String> = solutions
			.iter()
			.map(|expr| format!("x = {}", print_term(expr)))
			.collect();
		println!("Solutions: {{ {} }}", solutions_str.join(", "));

		println!("");
	}
}

pub fn solve(eq: &str, verbose: bool) -> Vec<RecExpr<EquationLanguage>> {
	let mut start = parse_equation(eq).unwrap();
	let root_id = Id::from(start.as_ref().len()-1);
	let EquationLanguage::Equals([left, right]) = start[root_id]
	else { panic!("Not an equation without an equals sign!"); };
	start[root_id] = EquationLanguage::Sub([left, right]);

	if verbose {
		println!("Parsed: {}", &start);
	}

	let runner = Runner::default()
		.with_explanations_enabled()
		.with_expr(&start)
		.run(&*RULES);

	let Some(factorization) = extract_normal_form(&runner.egraph, runner.roots[0]) else {
		panic!("Couldn't factorize polynomial!");
	};

	if verbose {
		println!("Factorized normal form: {}", factorization);
	}

	let mut solutions: Vec<RecExpr<EquationLanguage>> = Vec::new();
	for poly in &factorization.polynomials {
		if poly.len() == 2 { // linear factor
			let Rational { num, denom } = &poly[0];
			let mut solexpr = RecExpr::default();
			solexpr.add(EquationLanguage::Num(Rational {
				num: -*num,
				denom: *denom
			}));
			solutions.push(solexpr);
		} else if poly.len() == 3 { // quadratic factor
			let Rational { num: num0, denom: denom0 } = &poly[0];
			let Rational { num: num1, denom: denom1 } = &poly[1];

			let sol1 = format!("- ({num1})/(2 * ({denom1})) + ((({num1})/(2 * ({denom1}))) ^ 2 - ({num0}) / ({denom0})) ^ (1/2)");
			let sol2 = format!("- ({num1})/(2 * ({denom1})) - ((({num1})/(2 * ({denom1}))) ^ 2 - ({num0}) / ({denom0})) ^ (1/2)");

			let expr = parse_equation(&sol1).unwrap();
			let runner = Runner::default()
				.with_expr(&expr)
				.run(&*RULES);
			let extractor = Extractor::new(&runner.egraph, AstSize);
			let (_, simplified_expr) = extractor.find_best(runner.roots[0]);
			solutions.push(simplified_expr);

			let expr = parse_equation(&sol2).unwrap();
			let runner = Runner::default()
				.with_expr(&expr)
				.run(&*RULES);
			let extractor = Extractor::new(&runner.egraph, AstSize);
			let (_, simplified_expr) = extractor.find_best(runner.roots[0]);
			solutions.push(simplified_expr);
		} else if poly.len() == 4 {
			let guesses = approximate_rational_cubic(
				&Rational { num: 0, denom: 1 },
				&Rational { num: -91, denom: 1 },
				&Rational { num: -90, denom: 1 },
				1000
			);

			if guesses.len() < 3 {
				continue;
			}

			if verbose {
				println!("Guessing rational solutions: {}, {}, {}",
						 guesses[0], guesses[1], guesses[2]);
			}

			let start = format!("x ^ 3 + ({}) * x ^ 2 + ({}) * x + ({})",
								poly[2], poly[1], poly[0]);
			let goal = format!("(x - ({})) * (x - ({})) * (x - ({}))",
							   guesses[0], guesses[1], guesses[2]);

			let start_expr = parse_equation(&start).unwrap();
			let goal_expr = parse_equation(&goal).unwrap();

			let mut egraph: EGraph = Default::default();
			egraph.add_expr(&start_expr);
			egraph.add_expr(&goal_expr);

			let runner = Runner::default()
				.with_egraph(egraph)
				.with_node_limit(100_000)
				.with_iter_limit(100)
				.run(&*RULES);

			let equivs = runner.egraph.equivs(&start_expr, &goal_expr);

			if !equivs.is_empty() {
				if verbose {
					println!("Verified guessed solutions!");
				}

				// expressions are equivalent, so the guesses are actually solutions
				for s in &guesses {
					let mut solexpr = RecExpr::default();
					solexpr.add(EquationLanguage::Num(s.clone()));
					solutions.push(solexpr);
				}
			} else {
				if verbose {
					println!("Couldn't verify guessed solutions.");
				}
			}
		}
	}

	solutions
}
